//roommesh.c

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <string.h>
#include <crtdbg.h>
#include "types.h"
#include "MyMemory.h"
#include "trdefs.h"
#include "trmdefs.h"
#include "roommesh.h"
#include "linklist.h"

#define ASSERT(e) _ASSERT(e)

vertexlist_t *vertexhead = NULL;
vertexlist_t *vertextail = NULL;
short vertexlistcount = 0;

poly4list_t *poly4head = NULL;
poly4list_t *poly4tail = NULL;
short poly4listcount = 0;

poly3list_t *poly3head = NULL;
poly3list_t *poly3tail = NULL;
short poly3listcount = 0;

short VertexAdd(rmvertex2_t *vertex)
{
    vertexlist_t *v;
    int x;
    x = 0;
    v = vertexhead;
    while (v)
    {
        if (memcmp(&v->vertex, vertex, sizeof(rmvertex2_t)) == 0)
        {
            return x;
        }
        x++;
        v = v->next;
    }
    NewListItem(vertex, sizeof(rmvertex2_t), (linklist_t **)&vertexhead, (linklist_t **)&vertextail);
    vertexlistcount++;
    return x;
}

void FreeVertexList(void)
{
    linklist_t *v, *v2;
#ifdef _DEBUG
    int c = 0;
#endif
    v = (linklist_t *)vertexhead;
    while (v)
    {
        v2 = v->next;
        FreeListItem(v, sizeof(rmvertex2_t));
        v = v2;
#ifdef _DEBUG
        c++;
#endif
    }
#ifdef _DEBUG
    ASSERT(c == vertexlistcount);
#endif
    vertexlistcount = 0;
    vertexhead = NULL;
    vertextail = NULL;
}

poly4_t *Poly4Add(void)
{
    poly4list_t *p;
    p = (poly4list_t *)NewListItem(NULL, sizeof(poly4_t), (linklist_t **)&poly4head, (linklist_t **)&poly4tail);
    poly4listcount++;
    return p->poly4;
}

void FreePoly4List(void)
{
    linklist_t *l, *l2;
#ifdef _DEBUG
    int c = 0;
#endif
    l = (linklist_t *)poly4head;
    while (l)
    {
        l2 = l->next;
        FreeListItem(l, sizeof(poly4_t));
        l = l2;
#ifdef _DEBUG
        c++;
#endif
    }
#ifdef _DEBUG
    ASSERT(c == poly4listcount);
#endif
    poly4listcount = 0;
    poly4head = NULL;
    poly4tail = NULL;
}

poly3_t *Poly3Add(void)
{
    linklist_t *l;
    l = NewListItem(NULL, sizeof(poly3_t), (linklist_t **)&poly3head, (linklist_t **)&poly3tail);
    poly3listcount++;
    return (poly3_t *)l->data;
}

void FreePoly3List(void)
{
    linklist_t *l, *l2;
#ifdef _DEBUG
    int c = 0;
#endif
    l = (linklist_t *)poly3head;
    while (l)
    {
        l2 = l->next;
        FreeListItem(l, sizeof(poly3_t));
        l = l2;
#ifdef _DEBUG
        c++;
#endif
    }
#ifdef _DEBUG
    ASSERT(c == poly3listcount);
#endif
    poly3listcount = 0;
    poly3head = NULL;
    poly3tail = NULL;
}

poly4_t *BuildPoly4(rmvertex2_t *vertices, int first, long cw)
{
    poly4_t *poly4;
    int pos = first;
    poly4 = Poly4Add();
    if (cw)
    {
        poly4->vertices[pos++] = VertexAdd(&vertices[0]);
        pos %= 4;
        poly4->vertices[pos++] = VertexAdd(&vertices[1]);
        pos %= 4;
        poly4->vertices[pos++] = VertexAdd(&vertices[2]);
        pos %= 4;
        poly4->vertices[pos] = VertexAdd(&vertices[3]);
    }
    else
    {
        poly4->vertices[pos++] = VertexAdd(&vertices[3]);
        pos %= 4;
        poly4->vertices[pos++] = VertexAdd(&vertices[2]);
        pos %= 4;
        poly4->vertices[pos++] = VertexAdd(&vertices[1]);
        pos %= 4;
        poly4->vertices[pos] = VertexAdd(&vertices[0]);
    }
    poly4->texture = 0;
    return poly4;
}

poly3_t *BuildPoly3(rmvertex2_t *vertices, int first, long cw)
{
    poly3_t *poly3;
    int pos = first;
    poly3 = Poly3Add();
    if (cw)
    {
        poly3->vertices[pos++] = VertexAdd(&vertices[0]);
        pos %= 3;
        //poly3->vertices[1] = VertexAdd(&vertices[1]);
        poly3->vertices[pos++] = VertexAdd(&vertices[2]);
        pos %= 3;
        poly3->vertices[pos] = VertexAdd(&vertices[3]);
    }
    else
    {
        poly3->vertices[pos++] = VertexAdd(&vertices[3]);
        pos %= 3;
        //poly3->vertices[1] = VertexAdd(&vertices[1]);
        poly3->vertices[pos++] = VertexAdd(&vertices[2]);
        pos %= 3;
        poly3->vertices[pos] = VertexAdd(&vertices[0]);
    }
    poly3->texture = 0;
    return poly3;
}

long BuildRoomMesh(trm_room_t *room)
{
    rmvertex2_t vertices[4];
    trm_sector_t *sectors, *sector, *left, *right, *up, *down;
    rmvertex2_t *pvertex;
    poly4_t *poly4;
    poly3_t *poly3;
    vertexlist_t *v;
    poly4list_t *p4;
    poly3list_t *p3;
    u_char *finalmesh;
    long bNormal;
    int x, y;
    short numxsectors, numysectors;
    numxsectors = room->numxsectors;
    numysectors = room->numysectors;
    sectors = room->sectors;
    
    memset(vertices, 0, sizeof(vertices));

    //floor & ceiling
    for (x = 0; x < numxsectors; x++)
    {
        for (y = 0; y < numysectors; y++)
        {
            sector = &sectors[(x*numysectors)+y];
            if ((sector->floor != SECTOR_WALL) || (sector->ceiling != SECTOR_WALL))
            {
                //floor and ceiling
                bNormal = TRUE;
                
                /*
                    vertex order:
                    poly4:       poly3: (1 unused)
                     1  2           2
                                     
                     0  3        0  3
                */
                vertices[0].v.x = vertices[1].v.x = x*0x400;
                vertices[2].v.x = vertices[3].v.x = (x+1)*0x400;
                vertices[0].v.y = vertices[1].v.y = vertices[2].v.y = vertices[3].v.y = sector->floor*0x100;
                vertices[0].v.z = vertices[3].v.z = y*0x400;
                vertices[1].v.z = vertices[2].v.z = (y+1)*0x400;
                
                if (sector->lrfloor != 0)
                {
                    if (sector->lrfloor > 0)
                    {
                        //left side down
                        vertices[0].v.y += sector->lrfloor*0x100;
                        vertices[1].v.y += sector->lrfloor*0x100;
                    }
                    else if (sector->lrfloor < 0)
                    {
                        //right side down
                        vertices[2].v.y += -sector->lrfloor*0x100;
                        vertices[3].v.y += -sector->lrfloor*0x100;
                    }
                }

                if (sector->udfloor != 0)
                {
                    if (sector->udfloor > 0)
                    {
                        //bottom side down
                        vertices[0].v.y += sector->udfloor*0x100;
                        vertices[3].v.y += sector->udfloor*0x100;
                    }
                    else if (sector->udfloor < 0)
                    {
                        //top side down
                        vertices[1].v.y += -sector->udfloor*0x100;
                        vertices[2].v.y += -sector->udfloor*0x100;
                    }
                }

                BuildPoly4(vertices, 0, TRUE);
                
                vertices[0].v.y = vertices[1].v.y = vertices[2].v.y = vertices[3].v.y = sector->ceiling*0x100;

                if (sector->lrceiling != 0)
                {
                    if (sector->lrceiling > 0)
                    {
                        //left side down
                        vertices[0].v.y += sector->lrceiling*0x100;
                        vertices[1].v.y += sector->lrceiling*0x100;
                    }
                    else if (sector->lrceiling < 0)
                    {
                        //right side down
                        vertices[2].v.y += -sector->lrceiling*0x100;
                        vertices[3].v.y += -sector->lrceiling*0x100;
                    }
                }

                if (sector->udceiling != 0)
                {
                    if (sector->udceiling > 0)
                    {
                        //bottom side down
                        vertices[0].v.y += sector->udceiling*0x100;
                        vertices[3].v.y += sector->udceiling*0x100;
                    }
                    else if (sector->udceiling < 0)
                    {
                        //top side down
                        vertices[1].v.y += -sector->udceiling*0x100;
                        vertices[2].v.y += -sector->udceiling*0x100;
                    }
                }

                BuildPoly4(vertices, 0, FALSE);
            }
            else
            {
                bNormal = FALSE;
            }

            left = right = up = down = NULL;
            if (x == 0)
            {
                right = sector+numysectors;
            }
            else if (x == numxsectors-1)
            {
                left = sector-numysectors;
            }
            else
            {
                left = sector-numysectors;
                right = sector+numysectors;
            }

            if (y == 0)
            {
                up = sector+1;
            }
            else if (y == numysectors-1)
            {
                down = sector-1;
            }
            else
            {
                up = sector+1;
                down = sector-1;
            }

            if (left)
            {
                if (bNormal)
                {
                    if ((left->floor != SECTOR_WALL) || (left->ceiling != SECTOR_WALL))
                    {
                        //if left sector is lower, create lower left side
                        if (left->floor > sector->floor)
                        {
                            vertices[0].v.x = vertices[1].v.x = vertices[2].v.x = vertices[3].v.x = x*0x400;
                            vertices[0].v.y = vertices[3].v.y = left->floor*0x100;
                            vertices[1].v.y = vertices[2].v.y = sector->floor*0x100;
                            vertices[0].v.z = vertices[1].v.z = (y+1)*0x400;
                            vertices[2].v.z = vertices[3].v.z = y*0x400;

                            /*if (sector->lrfloor == 0)
                            {
                                //no slant
                                //vertices[0].v.y = vertices[1].v.y = vertices[2].v.y = vertices[3].v.y = sector->floor*0x100;
                                
                                vertices[1].v.y = vertices[2].v.y = sector->floor*0x100;
                            }
                            else if (sector->lrfloor > 0)
                            {
                                //left side down
                                //vertices[0].v.y = vertices[1].v.y = (sector->floor+sector->lrfloor)*0x100;
                                //vertices[2].v.y = vertices[3].v.y = sector->floor*0x100;
                                
                                //vertices[1].v.y = vertices[2].v.y = (sector->floor+sector->lrfloor)*0x100;
                            }
                            else if (sector->lrfloor < 0)
                            {
                                //right side down
                            }*/

                            BuildPoly4(vertices, 0, TRUE);
                        }
                        //if left ceiling is higher, create upper left side
                        if (left->ceiling < sector->ceiling)
                        {
                            vertices[0].v.x = vertices[1].v.x = vertices[2].v.x = vertices[3].v.x = x*0x400;
                            vertices[0].v.y = vertices[3].v.y = left->ceiling*0x100;
                            vertices[1].v.y = vertices[2].v.y = sector->ceiling*0x100;
                            vertices[0].v.z = vertices[1].v.z = (y+1)*0x400;
                            vertices[2].v.z = vertices[3].v.z = y*0x400;
                            BuildPoly4(vertices, 0, FALSE);
                        }
                    }
                }
                else
                {
                    //if left sector is normal, create entire left side
                    if ((left->floor != SECTOR_WALL) || (left->ceiling != SECTOR_WALL))
                    {
                        vertices[0].v.x = vertices[1].v.x = vertices[2].v.x = vertices[3].v.x = x*0x400;
                        vertices[1].v.y = vertices[2].v.y = left->ceiling*0x100;
                        vertices[0].v.z = vertices[1].v.z = (y+1)*0x400;
                        vertices[2].v.z = vertices[3].v.z = y*0x400;
                        if (left->lrfloor < 0)
                        {
                            vertices[0].v.y = vertices[3].v.y = (left->floor-left->lrfloor)*0x100;
                        }
                        else
                        {
                            vertices[0].v.y = vertices[3].v.y = left->floor*0x100;
                        }
                        BuildPoly4(vertices, 0, TRUE);
                        if (left->udfloor != 0)
                        {
                            vertices[0].v.x = vertices[1].v.x = vertices[2].v.x = vertices[3].v.x = x*0x400;
                            vertices[0].v.y = vertices[3].v.y = (left->floor-left->lrfloor)*0x100;
                            vertices[0].v.z = y*0x400;
                            vertices[2].v.z = vertices[3].v.z = (y+1)*0x400;
                            if (left->udfloor > 0)
                            {
                            }
                            else if (left->udfloor < 0)
                            {
                                //vertices[2].v.y = (left->floor-left->udfloor)*0x100;
                                vertices[2].v.y = (left->floor-left->lrfloor-left->udfloor)*0x100;
                            }
                            BuildPoly3(vertices, 2, TRUE);
                        }
                    }
                }
            }
            if (right)
            {
                if (bNormal)
                {
                    if ((right->floor != SECTOR_WALL) || (right->ceiling != SECTOR_WALL))
                    {
                        //if right sector is lower, create lower right side
                        if (right->floor > sector->floor)
                        {
                            vertices[0].v.x = vertices[1].v.x = vertices[2].v.x = vertices[3].v.x = (x+1)*0x400;
                            vertices[0].v.y = vertices[3].v.y = right->floor*0x100;
                            vertices[1].v.y = vertices[2].v.y = sector->floor*0x100;
                            vertices[0].v.z = vertices[1].v.z = (y+1)*0x400;
                            vertices[2].v.z = vertices[3].v.z = y*0x400;
                            BuildPoly4(vertices, 0, FALSE);
                        }
                        //if right ceiling is higher, create upper right side
                        if (right->ceiling < sector->ceiling)
                        {
                            vertices[0].v.x = vertices[1].v.x = vertices[2].v.x = vertices[3].v.x = (x+1)*0x400;
                            vertices[0].v.y = vertices[3].v.y = right->ceiling*0x100;
                            vertices[1].v.y = vertices[2].v.y = sector->ceiling*0x100;
                            vertices[0].v.z = vertices[1].v.z = (y+1)*0x400;
                            vertices[2].v.z = vertices[3].v.z = y*0x400;
                            BuildPoly4(vertices, 0, TRUE);
                        }
                    }
                }
                else
                {
                    //if right sector is normal, create entire right side
                    if ((right->floor != SECTOR_WALL) || (right->ceiling != SECTOR_WALL))
                    {
                        vertices[0].v.x = vertices[1].v.x = vertices[2].v.x = vertices[3].v.x = (x+1)*0x400;
                        vertices[0].v.y = vertices[3].v.y = right->floor*0x100;
                        vertices[1].v.y = vertices[2].v.y = right->ceiling*0x100;
                        vertices[0].v.z = vertices[1].v.z = (y+1)*0x400;
                        vertices[2].v.z = vertices[3].v.z = y*0x400;
                        BuildPoly4(vertices, 0, FALSE);
                    }
                }
            }
            if (up)
            {
                if (bNormal)
                {
                    if ((up->floor != SECTOR_WALL) || (up->ceiling != SECTOR_WALL))
                    {
                        //if up sector is lower, create lower up side
                        if (up->floor > sector->floor)
                        {
                            vertices[0].v.x = vertices[1].v.x = x*0x400;
                            vertices[2].v.x = vertices[3].v.x = (x+1)*0x400;
                            vertices[0].v.y = vertices[3].v.y = up->floor*0x100;
                            vertices[1].v.y = vertices[2].v.y = sector->floor*0x100;
                            vertices[0].v.z = vertices[1].v.z = vertices[2].v.z = vertices[3].v.z = (y+1)*0x400;
                            BuildPoly4(vertices, 0, FALSE);
                        }
                        //if up ceiling is higher, create upper up side
                        if (up->ceiling < sector->ceiling)
                        {
                            vertices[0].v.x = vertices[1].v.x = x*0x400;
                            vertices[2].v.x = vertices[3].v.x = (x+1)*0x400;
                            vertices[0].v.y = vertices[3].v.y = up->ceiling*0x100;
                            vertices[1].v.y = vertices[2].v.y = sector->ceiling*0x100;
                            vertices[0].v.z = vertices[1].v.z = vertices[2].v.z = vertices[3].v.z = (y+1)*0x400;
                            BuildPoly4(vertices, 0, TRUE);
                        }
                    }
                }
                else
                {
                    //if up sector is normal, create entire up side
                    if ((up->floor != SECTOR_WALL) || (up->ceiling != SECTOR_WALL))
                    {
                        vertices[0].v.x = vertices[1].v.x = x*0x400;
                        vertices[2].v.x = vertices[3].v.x = (x+1)*0x400;
                        vertices[0].v.y = vertices[3].v.y = up->floor*0x100;
                        vertices[1].v.y = vertices[2].v.y = up->ceiling*0x100;
                        vertices[0].v.z = vertices[1].v.z = vertices[2].v.z = vertices[3].v.z = (y+1)*0x400;
                        BuildPoly4(vertices, 0, FALSE);
                    }
                }
            }
            if (down)
            {
                if (bNormal)
                {
                    if ((down->floor != SECTOR_WALL) || (down->ceiling != SECTOR_WALL))
                    {
                        //if down sector is lower, create lower down side
                        if (down->floor > sector->floor)
                        {
                            vertices[0].v.x = vertices[1].v.x = x*0x400;
                            vertices[2].v.x = vertices[3].v.x = (x+1)*0x400;
                            vertices[0].v.y = vertices[3].v.y = down->floor*0x100;
                            vertices[1].v.y = vertices[2].v.y = sector->floor*0x100;
                            //vertices[2].v.y = sector->floor*0x100;
                            vertices[0].v.z = vertices[1].v.z = vertices[2].v.z = vertices[3].v.z = y*0x400;
                            
                            if (sector->lrfloor == 0)
                            {
                                //no slant
                                //vertices[0].v.y = vertices[1].v.y = vertices[2].v.y = vertices[3].v.y = sector->floor*0x100;
                                
                                vertices[1].v.y = vertices[2].v.y = sector->floor*0x100;
                                BuildPoly4(vertices, 0, TRUE);
                            }
                            else if (sector->lrfloor > 0)
                            {
                                //left side down
                                //vertices[0].v.y = vertices[1].v.y = (sector->floor+sector->lrfloor)*0x100;
                                //vertices[2].v.y = vertices[3].v.y = sector->floor*0x100;
                                if (sector->floor+sector->lrfloor == down->floor)
                                {
                                    vertices[1].v.y = (sector->floor+sector->lrfloor)*0x100;
                                    BuildPoly3(vertices, 0, TRUE);
                                }
                                else
                                {
                                    ASSERT(FALSE);
                                }

                                //vertices[1].v.y = (sector->floor+sector->lrfloor)*0x100;
                            }
                            else if (sector->lrfloor < 0)
                            {
                                //right side down
                            }
                        }
                        //if down ceiling is higher, create downper down side
                        if (down->ceiling < sector->ceiling)
                        {
                            vertices[0].v.x = vertices[1].v.x = x*0x400;
                            vertices[2].v.x = vertices[3].v.x = (x+1)*0x400;
                            vertices[0].v.y = vertices[3].v.y = down->ceiling*0x100;
                            vertices[1].v.y = vertices[2].v.y = sector->ceiling*0x100;
                            vertices[0].v.z = vertices[1].v.z = vertices[2].v.z = vertices[3].v.z = y*0x400;
                            BuildPoly4(vertices, 0, FALSE);
                        }
                    }
                }
                else
                {
                    //if down sector is normal, create entire down side
                    if ((down->floor != SECTOR_WALL) || (down->ceiling != SECTOR_WALL))
                    {
                        vertices[0].v.x = vertices[1].v.x = x*0x400;
                        vertices[2].v.x = vertices[3].v.x = (x+1)*0x400;
                        vertices[0].v.y = vertices[3].v.y = down->floor*0x100;
                        vertices[1].v.y = vertices[2].v.y = down->ceiling*0x100;
                        vertices[0].v.z = vertices[1].v.z = vertices[2].v.z = vertices[3].v.z = y*0x400;
                        BuildPoly4(vertices, 0, TRUE);
                    }
                }
            }
        }
    }
    room->mesh.numvertices = vertexlistcount;
    room->mesh.numpoly4s = poly4listcount;
    room->mesh.numpoly3s = poly3listcount;
    room->mesh.un1 = 0;
    room->nummeshwords = (2+(room->mesh.numvertices*sizeof(rmvertex2_t))+2+(room->mesh.numpoly4s*sizeof(poly4_t))+2+(room->mesh.numpoly3s*sizeof(poly3_t))+2)/2;
    finalmesh = (u_char *)GetMemBlock((room->mesh.numvertices*sizeof(rmvertex2_t))+(room->mesh.numpoly4s*sizeof(poly4_t))+(room->mesh.numpoly3s*sizeof(poly3_t)));
    room->mesh.vertices = pvertex = (rmvertex2_t *)finalmesh;
    v = vertexhead;
    while (v)
    {
        memcpy(pvertex, v->vertex, sizeof(rmvertex2_t));
        v = v->next;
        pvertex++;
    }
    room->mesh.poly4s = poly4 = (poly4_t *)pvertex;
    p4 = poly4head;
    while (p4)
    {
        memcpy(poly4, p4->poly4, sizeof(poly4_t));
        p4 = p4->next;
        poly4++;
    }
    room->mesh.poly3s = poly3 = (poly3_t *)poly4;
    p3 = poly3head;
    while (p3)
    {
        memcpy(poly3, p3->poly3, sizeof(poly3_t));
        p3 = p3->next;
        poly3++;
    }
    FreePoly4List();
    FreePoly3List();
    FreeVertexList();
    return TRUE;
}

#ifdef __cplusplus
}
#endif
