//MyMemory.c

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <crtdbg.h>
#include "types.h"
#include "MyMemory.h"

#define ASSERT(e) _ASSERT(e)

long bMyMemoryInitialized = FALSE;
long cbTotalMainMemory = 0;
long cbMainMemory = 0;

void *pMainMemory = NULL;
void *pMainHead = NULL;
memblock_t *pFreeBlocks = NULL;
long nTotalFreeBlocks = 0;
long nFreeBlocks = 0;
long nEndFreeBlock = 0;

long InitMyMemory(long totalmainmemory, long totalfreeblocks)
{
    if (pMainMemory != NULL)
    {
        ASSERT(FALSE);
    }
    if (cbTotalMainMemory != 0)
    {    
        ASSERT(FALSE);
    }
    bMyMemoryInitialized = FALSE;
    pMainHead = NULL;
    cbTotalMainMemory = 0;
    cbMainMemory = 0;
    pFreeBlocks = NULL;
    nTotalFreeBlocks = 0;
    nFreeBlocks = 0;
    nEndFreeBlock = 0;

    pFreeBlocks = (memblock_t *)malloc(totalfreeblocks*sizeof(memblock_t));
    if (pFreeBlocks == NULL)
    {    
        return FALSE;
    }
    nTotalFreeBlocks = totalfreeblocks;
    memset(pFreeBlocks, 0, nTotalFreeBlocks*sizeof(memblock_t));

    pMainMemory = malloc(totalmainmemory);
    if (pMainMemory == NULL)
    {
        free(pFreeBlocks);
        return FALSE;
    }
    pMainHead = pMainMemory;
    cbTotalMainMemory = totalmainmemory;
    cbMainMemory = cbTotalMainMemory;

    bMyMemoryInitialized = TRUE;
    return TRUE;
}

void FinishMyMemory(void)
{
    if (pFreeBlocks != NULL)
    {    
        free(pFreeBlocks);
    }
    if (pMainMemory != NULL)
    {    
        free(pMainMemory);
    }
    bMyMemoryInitialized = FALSE;
}

void *GetMemBlock(long size)
{
    void *memptr;
    int x;
    if (nFreeBlocks > 0)
    {
        for (x = 0; x < nEndFreeBlock; x++)
        {
            if (pFreeBlocks[x].memblock == NULL)
            {    
                continue;
            }
            if (pFreeBlocks[x].size >= size)
            {
                memptr = pFreeBlocks[x].memblock;
                if (pFreeBlocks[x].size > size)
                {
                    pFreeBlocks[x].memblock = ((char *)pFreeBlocks[x].memblock)+size;
                    pFreeBlocks[x].size -= size;
                }
                else
                {
                    pFreeBlocks[x].memblock = NULL;
                    pFreeBlocks[x].size = 0;
                    nFreeBlocks--;
                    if (nFreeBlocks == 0)
                    {    
                        nEndFreeBlock = 0;
                    }
                    else if (x+1 == nEndFreeBlock)
                    {    
                        nEndFreeBlock--;
                    }
                }
                return(memptr);
            }
        }
    }
    if (cbMainMemory >= size)
    {
        memptr = pMainHead;
        pMainHead = ((char *)pMainHead)+size;
        cbMainMemory -= size;
        return(memptr);
    }
    else
    {
        //out of memory
        ASSERT(FALSE);
    }
    return NULL;
}

void FreeMemBlock(void *memblock, long size)
{
    int x;
    for (x = 0; x < nEndFreeBlock; x++)
    {
        if (pFreeBlocks[x].memblock != NULL)
        {    
            continue;
        }
        pFreeBlocks[x].memblock = memblock;
        pFreeBlocks[x].size = size;
        nFreeBlocks++;
        return;
    }
    pFreeBlocks[x].memblock = memblock;
    pFreeBlocks[x].size = size;
    nFreeBlocks++;
    nEndFreeBlock++;
}

void FreeAllBlocks(void)
{
    cbMainMemory = cbTotalMainMemory;
    pMainHead = pMainMemory;
    nFreeBlocks = 0;
    nEndFreeBlock = 0;
    memset(pFreeBlocks, 0, nTotalFreeBlocks*sizeof(memblock_t));
    memset(pMainMemory, 0, cbTotalMainMemory);
}

#ifdef __cplusplus
}
#endif
