//trdefs.h

#ifndef _TRDEFS_INCLUDED_
#define _TRDEFS_INCLUDED_

#define TR2_VERSION 0x2d
#define TR2_NUMCOLORMAPS 32

#define SECTOR_WALL -127

typedef unsigned char u_char;
typedef unsigned short u_short;
typedef unsigned long u_long;
typedef signed char s_char;

typedef struct
{
    short x;
    short y;
    short z;
} vector_t;

typedef struct
{
    short vertices[4]; //indexes to some vertices
    short texture; //object texture #
} face4_t; //rectangle

typedef struct
{
    short vertices[4]; //indexes to some vertices
    short texture; //object texture #
} poly4_t; //rectangle

typedef struct
{
    short vertices[3]; //indexes to some vertices
    short texture; //object texture #
} poly3_t; //triangle

typedef struct
{
    short vertices[3]; //indexes to some vertices
    short texture; //object texture #
} face3_t; //triangle

typedef struct
{
    u_char tile[256*256];
} textile8_t;

typedef struct
{
    u_char tile[(256*256)*2];
} textile16_t;

typedef struct
{
    long x;
    long z;
    u_char un1[8];
} roominfo_t;

typedef struct
{
    short room;
    vector_t normal; //?
    vector_t vertices[4];
} roomdoor_t;

typedef struct
{
    short fdindex; //slope of ceiling and floor? index to slopes?
    short camera;
    s_char floorattrib; //if 0xff floor is solid, if 0x0e, floor is not solid?, 0x01, 
    s_char floor;
    s_char ceilingattrib;
    s_char ceiling;
} sector_t;

typedef struct
{
    u_char un1[24];
} roomlight2_t;

typedef struct
{
    long offset;
    roominfo_t roominfo;
    long numrtwords; //size in words of the room mesh, sprite data for this room
    u_char *rtwords;
    /*short numvertices;
    rmvertex_t *vertices;
    short numrectangles;
    face4_t *rectangles;
    short numtriangles;
    face3_t *triangles;
    short numsprites;
    roomsprite_t *sprites;*/
    short numdoors;
    roomdoor_t *doors;
    short numysectors;
    short numxsectors;
    sector_t *sectors;
    u_char un1;
    u_char lightlevel1;
    u_char un4;
    u_char lightlevel2;
    short numlights;
    roomlight2_t *lights;
    //short numstatics;
    //roomstatic_t *statics;
    short un2; //always 0xffff?
    u_char water; //bit 0 set = water
    u_char un3;
} room2_t;

typedef struct
{
    vector_t v;
    u_char un1;
    u_char colormap; //colormap # for this vertex
} rmvertex_t; //room mesh vertex

typedef struct
{
    vector_t v;
    //u_char un1;
    //u_char colormap; //colormap # for this vertex
    short un1;
    short un2;
    short un3;
} rmvertex2_t; //room mesh vertex

typedef struct
{
    short vertex;
    short sprite;
} roomsprite_t;

typedef struct
{
    u_char un1[18];
} roomstatic_t;

typedef struct
{
    u_char un1[10];
    short numvectors;
    vector_t *vectors;
    short numun2;
    vector_t *un2;
    short numrectangles;
    face4_t *rectangles;
    //more
} mesh_t;

typedef struct
{
    short un1; // may not be shorts?
    short tile;
    
    u_char un2;
    u_char ulx;
    
    u_char un3;
    u_char uly;
    
    u_char un4;
    u_char urx;
    
    u_char un5;
    u_char ury;
    
    u_char un6;
    u_char lrx;
    
    u_char un7;
    u_char lry;
    
    u_char un8;
    u_char llx;
    
    u_char un9;
    u_char lly;
} objtexture_t;

typedef struct
{
    u_char colormap[256];
} colormap_t;

typedef struct
{
    short sample;
    short volume;
    short un1;
    short un2;
} sampleinfo_t;

typedef struct
{
    short object;
    u_char un1[16];
    short light;
    u_char un2[2];
} item_t;

typedef struct
{
    short object;
    short room;
    long x;
    long y;
    long z;
    short angle;
    short light;
    short un2;
    short un3;
} item2_t;

typedef struct
{
    u_char un1[32];
} anim_t;

typedef struct
{
    u_char un1[6];
} struct_t;

typedef struct
{
    u_char un1[8];
} range_t;

typedef struct
{
    u_char un1[2];
} bone1_t;

typedef struct
{
    u_char un1[4];
} bone2_t;

typedef struct
{
    u_char un1[2];
} frame_t;

typedef struct
{
    u_char un1[16];
} spritetexture_t;

typedef struct
{
    u_char un1[16];
} camera_t;

typedef struct
{
    u_char un1[16];
} soundfx_t;

typedef struct
{
    u_char un1[20];
} box_t;

typedef struct
{
    u_char un1[2];
} overlap_t;

typedef struct
{
    u_char un1[2];
} animtexture_t;

typedef struct
{
    u_char un1[18];
} un2_t;

typedef struct
{
    u_char un1[32];
} un3_t;

typedef struct
{
    u_char un1[8];
} un4_t;

typedef struct
{
    u_char un1[12];
} un5_t;

typedef struct
{
    u_char un1[16];
} un6_t;

typedef struct
{
    long bIsLoaded;
    unsigned long dwLoadFlags;
    //offsets
    int nFileType;
} mytr2vars_t;

typedef struct
{
    long bIsLoaded;
    unsigned long dwLoadFlags;
    //offsets
    long roomoffset;
    long floordataoffset;
    long itemoffset;
    int nFileType;
} mytr1vars_t;

typedef struct
{
    u_char un1[18];
} roomlight_t;

typedef struct
{
    short numvertices;
    rmvertex2_t *vertices;
    short numpoly4s;
    poly4_t *poly4s;
    short numpoly3s;
    poly3_t *poly3s;
    short un1; //always 0?
} roommesh2_t;

typedef struct
{
    long offset;
    roominfo_t roominfo;
    long numrtwords; //size in words of the room mesh, sprite data for this room
    u_char *rtwords;
    short numvertices;
    rmvertex_t *vertices;
    short numrectangles;
    face4_t *rectangles;
    short numtriangles;
    face3_t *triangles;
    short numsprites;
    roomsprite_t *sprites;
    short numdoors;
    roomdoor_t *doors;
    short numysectors;
    short numxsectors;
    sector_t *sectors;
    u_char un1;
    u_char lightlevel;
    short numlights;
    roomlight_t *lights;
    short numstatics;
    roomstatic_t *statics;
    short un2; //always 0xffff?
    u_char water; //bit 0 set = water
    u_char un3;
} room_t;

typedef struct
{
    long palette;
    long textiles8;
    long textiles16;
    long rooms;
    long roomsize;
    long floordata;
    long meshes;
    long meshpointers;
    long anims;
    long structs;
    long ranges;
    long bones1;
    long bones2;
    long frames;
    long un2;
    long un3;
    long objtextures;
    long spritetextures;
    long un4;
    long colormaps;
    long cameras;
    long soundfx;
    long boxes1;
    long overlaps;
    long boxes2;
    long animtextures;
    long items;
    long un6;
    long demodata;
    long soundstuff;
    long sampleinfos;
    long samples;
} tr2offsets_t;

typedef struct
{
    long version; //0x2a for tr2 levels
    u_char *palette;
    long numtextiles;
    textile8_t *textiles8;
    textile16_t *textiles16;
    long un1; //always 0?
    short numrooms;
    room2_t *rooms;
    long numfloordata;
    u_char *floordata;
    long nummeshwords;
    short *meshwords;
    //mesh_t *meshes;
    long nummeshpointers;
    long *meshpointers;
    long numobjtextures;
    objtexture_t *objtextures;
    long numitems;
    item2_t *items;
    long numanims;
    anim_t *anims;
    long numstructs;
    struct_t *structs;
    long numranges;
    range_t *ranges;
    long numbones1;
    bone1_t *bones1;
    long numbones2;
    bone2_t *bones2;
    long numframes;
    frame_t *frames;
    long numun2;
    un2_t *un2;
    long numun3;
    un3_t *un3;
    long numspritetextures;
    spritetexture_t *spritetextures;
    long numun4;
    un4_t *un4;
    long numcameras;
    camera_t *cameras;
    long numsoundfx;
    soundfx_t *soundfx;
    long numboxes;
    box_t *boxes;
    long numoverlaps;
    overlap_t *overlaps;
    un5_t *un5;
    long numanimtextures;
    animtexture_t *animtextures;
    short numun6;
    un6_t *un6;
    short demodatasize;
    u_char *demodata;
    u_char *soundstuff;
    long numsampleinfos;
    sampleinfo_t *sampleinfos;
    long samplesize;
    u_char *samples;
    long numsampleoffsets;
    long *sampleoffsets;
    colormap_t *colormaps;
    long numsamples;
    long *sampleindexes;
    tr2offsets_t offsets;
    mytr2vars_t myvars;
} tr2_t;

typedef struct
{
    long version; //0x20 for tr1 levels
    long numtextiles;
    textile8_t *textiles8;
    long un1; //always 0?
    short numrooms;
    room_t *rooms;
    long numfloordata;
    u_char *floordata;
    long nummeshwords;
    short *meshwords;
    //mesh_t *meshes;
    long nummeshpointers;
    long *meshpointers;
    long numobjtextures;
    objtexture_t *objtextures;
    u_char *palette;
    long numitems;
    item_t *items;
    long numanims;
    anim_t *anims;
    long numstructs;
    struct_t *structs;
    long numranges;
    range_t *ranges;
    long numbones1;
    bone1_t *bones1;
    long numbones2;
    bone2_t *bones2;
    long numframes;
    frame_t *frames;
    long numun2;
    un2_t *un2;
    long numun3;
    un3_t *un3;
    long numspritetextures;
    spritetexture_t *spritetextures;
    long numun4;
    un4_t *un4;
    long numcameras;
    camera_t *cameras;
    long numsoundfx;
    soundfx_t *soundfx;
    long numboxes;
    box_t *boxes;
    long numoverlaps;
    overlap_t *overlaps;
    un5_t *un5;
    long numanimtextures;
    animtexture_t *animtextures;
    short numun6;
    un6_t *un6;
    short demodatasize;
    u_char *demodata;
    u_char *soundstuff;
    long numsampleinfos;
    sampleinfo_t *sampleinfos;
    long samplesize;
    u_char *samples;
    long numsampleoffsets;
    long *sampleoffsets;
    colormap_t *colormaps;
    mytr1vars_t myvars;
} PHD_t;

#endif //_TRDEFS_INCLUDED_
